from typing import Any, Callable, Dict, Iterator, List, Optional, Union
from IPython.display import HTML, Image
import pydot

from collections import Counter
import json
import os
import re
from copy import deepcopy
from pathlib import Path

import lunchbox.tools as lbt
from pandas import DataFrame
import networkx

import rolling_pin.tools as tools
# ------------------------------------------------------------------------------

'''
Contains the BlobETL class, which is used for coverting JSON blobs, and their
python equivalents, into flat dictionaries that can easily be modified and
converted to directed graphs.
'''


class BlobETL():
    '''
    Converts blob data internally into a flat dictionary that is universally
    searchable, editable and convertable back to the data's original structure,
    new blob structures or dircted graphs.
    '''
    def __init__(self, blob, separator='/'):
        # type: (Any, str) -> None
        '''
        Contructs BlobETL instance.

        Args:
            blob (object): Iterable object.
            separator (str, optional): String to be used as a field separator in
                each key. Default: '/'.
        '''
        self._data = tools \
            .flatten(blob, separator=separator, embed_types=True)  # type: Dict[str, Any]
        self._separator = separator  # type: str

    # EDIT_METHODS--------------------------------------------------------------
    def query(self, regex, ignore_case=True):
        # type: (str, bool) -> BlobETL
        '''
        Filter data items by key according to given regular expression.

        Args:
            regex (str): Regular expression.
            ignore_casd (bool, optional): Whether to consider case in the
                regular expression search. Default: False.

        Returns:
            BlobETL: New BlobETL instance.
        '''
        if ignore_case:
            return self.filter(lambda x: bool(re.search(regex, x, re.I)), by='key')
        return self.filter(lambda x: bool(re.search(regex, x)), by='key')

    def filter(self, predicate, by='key'):
        # type: (Callable[[Any], bool], str) -> BlobETL
        '''
        Filter data items by key, value or key + value, according to a given
        predicate.

        Args:
            predicate: Function that returns a boolean value.
            by (str, optional): Value handed to predicate.
                Options include: key, value, key+value. Default: key.

        Raises:
            ValueError: If by keyword is not key, value, or key+value.

        Returns:
            BlobETL: New BlobETL instance.
        '''
        data = {}
        if by not in ['key', 'value', 'key+value']:
            msg = f'Invalid by argument: {by}. Needs to be one of: '
            msg += 'key, value, key+value.'
            raise ValueError(msg)

        for key, val in self._data.items():
            item = None
            if by == 'key':
                item = [key]
            elif by == 'value':
                item = [val]
            else:
                item = [key, val]

            if predicate(*item):
                data[key] = val

        return BlobETL(data, separator=self._separator)

    def delete(self, predicate, by='key'):
        # type: (Callable[[Any], bool], str) -> BlobETL
        '''
        Delete data items by key, value or key + value, according to a given
        predicate.

        Args:
            predicate: Function that returns a boolean value.
            by (str, optional): Value handed to predicate.
                Options include: key, value, key+value. Default: key.

        Raises:
            ValueError: If by keyword is not key, value, or key+value.

        Returns:
            BlobETL: New BlobETL instance.
        '''
        data = deepcopy(self._data)
        if by not in ['key', 'value', 'key+value']:
            msg = f'Invalid by argument: {by}. Needs to be one of: '
            msg += 'key, value, key+value.'
            raise ValueError(msg)

        for key, val in self._data.items():
            item = None
            if by == 'key':
                item = [key]
            elif by == 'value':
                item = [val]
            else:
                item = [key, val]

            if predicate(*item):
                del data[key]

        return BlobETL(data, separator=self._separator)

    def set(
        self,
        predicate=None,  # type: Optional[Callable[[Any, Any], bool]]
        key_setter=None,  # type: Optional[Callable[[Any, Any], str]]
        value_setter=None,  # type: Optional[Callable[[Any, Any], Any]]
    ):
        # type: (...) -> BlobETL
        '''
        Filter data items by key, value or key + value, according to a given
        predicate. Then set that items key by a given function and value by a
        given function.

        Args:
            predicate (function, optional): Function of the form:
                lambda k, v: bool. Default: None --> lambda k, v: True.
            key_setter (function, optional): Function of the form:
                lambda k, v: str. Default: None --> lambda k, v: k.
            value_setter (function, optional):  Function of the form:
                lambda k, v: object. Default: None --> lambda k, v: v.

        Returns:
            BlobETL: New BlobETL instance.
        '''
        # assign default predicate
        if predicate is None:
            predicate = lambda k, v: True

        # assign default key_setter
        if key_setter is None:
            key_setter = lambda k, v: k

        # assign default value_setter
        if value_setter is None:
            value_setter = lambda k, v: v

        data = deepcopy(self._data)
        for item in self._data.items():
            if predicate(*item):
                k = key_setter(*item)
                v = value_setter(*item)
                del data[item[0]]
                data[k] = v

        return BlobETL(data, separator=self._separator)

    def update(self, item):
        # type: (Union[Dict, BlobETL]) -> BlobETL
        '''
        Updates internal dictionary with given dictionary or BlobETL instance.
        Given dictionary is first flattened with embeded types.

        Args:
            item (dict or BlobETL): Dictionary to be used for update.

        Returns:
            BlobETL: New BlobETL instance.
        '''
        if isinstance(item, BlobETL):
            item = item._data
        temp = tools.flatten(item, separator=self._separator, embed_types=True)
        data = deepcopy(self._data)
        data.update(temp)
        return BlobETL(data, separator=self._separator)

    def set_field(self, index, field_setter):
        # type: (int, Callable[[str], str]) -> BlobETL
        '''
        Set's a field at a given index according to a given function.

        Args:
            index (int): Field index.
            field_setter (function): Function of form lambda str: str.

        Returns:
            BlobETL: New BlobETL instance.
        '''
        output = {}
        for key, val in self._data.items():
            fields = key.split(self._separator)
            fields[index] = field_setter(fields[index])
            key = self._separator.join(fields)
            output[key] = val
        return BlobETL(output, separator=self._separator)

    # EXPORT-METHODS------------------------------------------------------------
    def to_dict(self):
        # type: () -> Dict[str, Any]
        '''
        Returns:
            dict: Nested representation of internal data.
        '''
        return tools.unembed(
            tools.nest(deepcopy(self._data), separator=self._separator)
        )

    def to_flat_dict(self):
        # type: () -> Dict[str, Any]
        '''
        Returns:
            dict: Flat dictionary with embedded types.
        '''
        return deepcopy(self._data)

    def to_records(self):
        # type: () -> List[Dict]
        '''
        Returns:
            list[dict]: Data in records format.
        '''
        data = []
        for key, val in self._data.items():
            fields = key.split(self._separator)
            row = {i: v for i, v in enumerate(fields)}  # type: Dict[Any, Any]
            row['value'] = val
            data.append(row)
        return data

    def to_dataframe(self, group_by=None):
        # type: (Optional[int]) -> DataFrame
        '''
        Convert data to pandas DataFrame.

        Args:
            group_by (int, optional): Field index to group rows of data by.
                Default: None.

        Returns:
            DataFrame: DataFrame.
        '''
        data = self.to_records()  # type: Any
        data = DataFrame(data)

        if group_by is not None:
            group = list(range(0, group_by))
            data = DataFrame(data)\
                .groupby(group, as_index=False)\
                .agg(lambda x: x.tolist())\
                .apply(lambda x: x.to_dict(), axis=1)\
                .tolist()
            data = DataFrame(data)

        # clean up column order
        cols = data.columns.tolist()  # type: List[str]
        cols = list(sorted(filter(lambda x: x != 'value', cols)))
        cols += ['value']
        data = data[cols]

        return data

    def to_prototype(self):
        # type: () -> BlobETL
        '''
        Convert data to prototypical representation.

        Example:

        >>> data = {
        'users': [
                {
                    'name': {
                        'first': 'tom',
                        'last': 'smith',
                    }
                },{
                    'name': {
                        'first': 'dick',
                        'last': 'smith',
                    }
                },{
                    'name': {
                        'first': 'jane',
                        'last': 'doe',
                    }
                },
            ]
        }
        >>> BlobETL(data).to_prototype().to_dict()
        {
            '^users': {
                '<list_[0-9]+>': {
                    'name': {
                        'first$': Counter({'dick': 1, 'jane': 1, 'tom': 1}),
                        'last$': Counter({'doe': 1, 'smith': 2})
                    }
                }
            }
        }

        Returns:
            BlobETL: New BlobETL instance.
        '''
        def regex_in_list(regex, items):
            # type: (str, List[str]) -> bool
            for item in items:
                if re.search(regex, item):
                    return True
            return False  # pragma: no cover

        def field_combinations(a, b):
            # type: (List[str], List[str]) -> List[str]
            output = []
            for fa in a:
                for fb in b:
                    output.append(fa + self._separator + fb)
            return output

        keys = list(self._data.keys())
        fields = list(map(lambda x: x.split(self._separator), keys))

        fields = DataFrame(fields)\
            .apply(lambda x: x.unique().tolist())\
            .apply(lambda x: filter(lambda y: y is not None, x)) \
            .apply(lambda x: map(
                lambda y: re.sub(r'<([a-z]+)_\d+>', '<\\1_[0-9]+>', y),
                x)) \
            .apply(lambda x: list(set(x))) \
            .tolist()

        prev = fields[0]
        regexes = list()
        for i, level in enumerate(fields[1:]):
            temp = field_combinations(prev, level)  # type: Union[List, Iterator]
            temp = filter(lambda x: regex_in_list('^' + x, keys), temp)
            prev = list(temp)
            regexes.extend(prev)

        regexes = lbt.get_ordered_unique(regexes)

        p_keys = set()
        for regex in regexes:
            other = deepcopy(regexes)
            other.remove(regex)
            not_in_other = True
            for item in other:
                if regex in item:
                    not_in_other = False
            if not_in_other:
                p_keys.add(f'^{regex}$')

        output = {}
        for key in p_keys:
            values = self.query(key).to_flat_dict().values()
            output[key] = Counter(values)
        return BlobETL(output, separator=self._separator)

    def to_networkx_graph(self):
        # type: () -> networkx.DiGraph
        '''
        Converts internal dictionary into a networkx directed graph.

        Returns:
            networkx.DiGraph: Graph representation of dictionary.
        '''
        graph = networkx.DiGraph()
        graph.add_node('root')
        embed_re = re.compile(r'<[a-z]+_(\d+)>')

        def recurse(item, parent):
            # type: (Dict, str) -> None
            for key, val in item.items():
                k = f'{parent}{self._separator}{key}'
                short_name = embed_re.sub('\\1', key)
                graph.add_node(k, short_name=short_name, node_type='key')
                graph.add_edge(parent, k)

                if isinstance(val, dict):
                    recurse(val, k)
                else:
                    graph.nodes[k]['value'] = [val]
                    name = f'"{str(val)}"'
                    v = f'"{k}{self._separator}{str(val)}"'
                    graph.add_node(
                        v, short_name=name, node_type='value', value=[val]
                    )
                    graph.add_edge(k, v)

        recurse(tools.nest(self._data, self._separator), 'root')
        graph.remove_node('root')
        return graph

    def to_dot_graph(
        self, orthogonal_edges=False, orient='tb', color_scheme=None
    ):
        # type: (bool, str, Optional[Dict[str, str]]) -> pydot.Dot
        '''
        Converts internal dictionary into pydot graph.
        Key and value nodes and edges are colored differently.

        Args:
            orthogonal_edges (bool, optional): Whether graph edges should have
                non-right angles. Default: False.
            orient (str, optional): Graph layout orientation. Default: tb.
                Options include:

                * tb - top to bottom
                * bt - bottom to top
                * lr - left to right
                * rl - right to left
            color_scheme: (dict, optional): Color scheme to be applied to graph.
                Default: rolling_pin.tools.COLOR_SCHEME

        Raises:
            ValueError: If orient is invalid.

        Returns:
            pydot.Dot: Dot graph representation of dictionary.
        '''
        orient = orient.lower()
        orientations = ['tb', 'bt', 'lr', 'rl']
        if orient not in orientations:
            msg = f'Invalid orient value. {orient} not in {orientations}.'
            raise ValueError(msg)

        # set default colort scheme
        if color_scheme is None:
            color_scheme = tools.COLOR_SCHEME

        # create pydot graph
        graph = self.to_networkx_graph()
        dot = networkx.drawing.nx_pydot.to_pydot(graph)

        # set layout orientation
        dot.set_rankdir(orient.upper())

        # set graph background color
        dot.set_bgcolor(color_scheme['background'])

        # set edge draw type
        if orthogonal_edges:
            dot.set_splines('ortho')

        # set draw parameters for each node of graph
        for node in dot.get_nodes():
            node.set_shape('rect')
            node.set_style('filled')
            node.set_color(color_scheme['node'])
            node.set_fillcolor(color_scheme['node'])
            node.set_fontcolor(color_scheme['node_font'])
            node.set_fontname('Courier')

            # if node has short name, set its displayed name to that
            attrs = node.get_attributes()
            if 'short_name' in attrs:
                node.set_label(attrs['short_name'])

            # if node type is value change its colors
            if 'node_type' in attrs and attrs['node_type'] == 'value':
                node.set_color(color_scheme['node_value'])
                node.set_fillcolor(color_scheme['node_value'])
                node.set_fontcolor(color_scheme['node_value_font'])

        # set draw parameters for each edge in graph
        for edge in dot.get_edges():
            edge.set_color(color_scheme['edge'])

            # if edge destination node type is value change its color
            node = dot.get_node(edge.get_destination())[0]
            attrs = node.get_attributes()
            if 'node_type' in attrs and attrs['node_type'] == 'value':
                edge.set_color(color_scheme['edge_value'])

        return dot

    def to_html(
        self,
        layout='dot',
        orthogonal_edges=False,
        orient='tb',
        color_scheme=None,
        as_png=False,
    ):
        # type: (str, bool, str, Optional[Dict[str, str]], bool) -> Union[Image, HTML]
        '''
        For use in inline rendering of graph data in Jupyter Lab.

        Args:
            layout (str, optional): Graph layout style.
                Options include: circo, dot, fdp, neato, sfdp, twopi.
                Default: dot.
            orthogonal_edges (bool, optional): Whether graph edges should have
                non-right angles. Default: False.
            orient (str, optional): Graph layout orientation. Default: tb.
                Options include:

                * tb - top to bottom
                * bt - bottom to top
                * lr - left to right
                * rl - right to left
            color_scheme: (dict, optional): Color scheme to be applied to graph.
                Default: rolling_pin.tools.COLOR_SCHEME
            as_png (bool, optional): Display graph as a PNG image instead of
                SVG. Useful for display on Github. Default: False.

        Returns:
            IPython.display.HTML: HTML object for inline display.
        '''
        if color_scheme is None:
            color_scheme = tools.COLOR_SCHEME

        dot = self.to_dot_graph(
            orthogonal_edges=orthogonal_edges,
            orient=orient,
            color_scheme=color_scheme,
        )
        return tools.dot_to_html(dot, layout=layout, as_png=as_png)

    def write(
        self,
        fullpath,
        layout='dot',
        orthogonal_edges=False,
        orient='tb',
        color_scheme=None
    ):
        # type: (Union[str, Path], str, bool, str, Dict[str, str]) -> BlobETL
        '''
        Writes internal dictionary to a given filepath.
        Formats supported: svg, dot, png, json.

        Args:
            fulllpath (str or Path): File tobe written to.
            layout (str, optional): Graph layout style.
                Options include: circo, dot, fdp, neato, sfdp, twopi.
                Default: dot.
            orthogonal_edges (bool, optional): Whether graph edges should have
                non-right angles. Default: False.
            orient (str, optional): Graph layout orientation. Default: tb.
                Options include:

                * tb - top to bottom
                * bt - bottom to top
                * lr - left to right
                * rl - right to left
            color_scheme: (dict, optional): Color scheme to be applied to graph.
                Default: rolling_pin.tools.COLOR_SCHEME

        Raises:
            ValueError: If invalid file extension given.

        Returns:
            BlobETL: self.
        '''
        if isinstance(fullpath, Path):
            fullpath = fullpath.absolute().as_posix()

        _, ext = os.path.splitext(fullpath)
        ext = re.sub(r'^\.', '', ext)
        if re.search('^json$', ext, re.I):
            with open(fullpath, 'w') as f:
                json.dump(self.to_dict(), f)
            return self

        if color_scheme is None:
            color_scheme = tools.COLOR_SCHEME

        graph = self.to_dot_graph(
            orthogonal_edges=orthogonal_edges,
            orient=orient,
            color_scheme=color_scheme,
        )
        try:
            tools.write_dot_graph(graph, fullpath, layout=layout,)
        except ValueError:
            msg = f'Invalid extension found: {ext}. '
            msg += 'Valid extensions include: svg, dot, png, json.'
            raise ValueError(msg)
        return self



from typing import Any, Dict, Iterable, List, Union
import pydot

import logging
import os
import re
from collections import OrderedDict
from pathlib import Path

from IPython.display import HTML, Image

LOG_LEVEL = os.environ.get('LOG_LEVEL', 'WARNING').upper()
logging.basicConfig(level=LOG_LEVEL)
LOGGER = logging.getLogger(__name__)
# ------------------------------------------------------------------------------

'''
Contains basic functions for more complex ETL functions and classes.
'''

# COLOR-SCHEME------------------------------------------------------------------
COLOR_SCHEME = dict(
    background='#242424',
    node='#343434',
    node_font='#B6ECF3',
    node_value='#343434',
    node_value_font='#DE958E',
    edge='#B6ECF3',
    edge_value='#DE958E',
    node_library_font='#DE958E',
    node_subpackage_font='#A0D17B',
    node_module_font='#B6ECF3',
    edge_library='#DE958E',
    edge_subpackage='#A0D17B',
    edge_module='#B6ECF3',
)  # type: Dict[str, str]

COLOR_SCALE = [
    '#B6ECF3',
    '#DE958E',
    '#EBB483',
    '#A0D17B',
    '#93B6E6',
    '#AC92DE',
    '#E9EABE',
    '#7EC4CF',
    '#F77E70',
    '#EB9E58',
]  # type: List[str]


# PREDICATE-FUNCTIONS-----------------------------------------------------------
def is_iterable(item):
    # type: (Any) -> bool
    '''
    Determines if given item is iterable.

    Args:
        item (object): Object to be tested.

    Returns:
        bool: Whether given item is iterable.
    '''
    if is_listlike(item) or is_dictlike(item):
        return True
    return False


def is_dictlike(item):
    # type: (Any) -> bool
    '''
    Determines if given item is dict-like.

    Args:
        item (object): Object to be tested.

    Returns:
        bool: Whether given item is dict-like.
    '''
    for type_ in [dict, OrderedDict]:
        if isinstance(item, type_):
            if item.__class__.__name__ == 'Counter':
                return False
            return True
    return False


def is_listlike(item):
    # type: (Any) -> bool
    '''
    Determines if given item is list-like.

    Args:
        item (object): Object to be tested.

    Returns:
        bool: Whether given item is list-like.
    '''
    for type_ in [list, tuple, set]:
        if isinstance(item, type_):
            return True
    return False


# CORE-FUNCTIONS----------------------------------------------------------------
def flatten(item, separator='/', embed_types=True):
    # type: (Iterable, str, bool) -> Dict[str, Any]
    '''
    Flattens a iterable object into a flat dictionary.

    Args:
        item (object): Iterable object.
        separator (str, optional): Field separator in keys. Default: '/'.

    Returns:
        dict: Dictionary representation of given object.
    '''
    output = {}  # type: Dict[str, Any]

    def recurse(item, cursor):
        # type (Iterable, Any) -> None
        if is_listlike(item):
            if embed_types:
                name = item.__class__.__name__
                item = [(f'<{name}_{i}>', val) for i, val in enumerate(item)]
                item = dict(item)
            else:
                item = dict(enumerate(item))
        if is_dictlike(item):
            for key, val in item.items():
                new_key = f'{cursor}{separator}{str(key)}'
                if is_iterable(val) and len(val) > 0:
                    recurse(val, new_key)
                else:
                    final_key = re.sub('^' + separator, '', new_key)
                    output[final_key] = val

    recurse(item, '')
    return output


def nest(flat_dict, separator='/'):
    # type: (Dict[str, Any], str) -> Dict[str, Any]
    '''
    Converts a flat dictionary into a nested dictionary by splitting keys by a
    given separator.

    Args:
        flat_dict (dict): Flat dictionary.
        separator (str, optional): Field separator within given dictionary's
            keys. Default: '/'.

    Returns:
        dict: Nested dictionary.
    '''
    output = {}  # type: Dict[str, Any]
    for keys, val in flat_dict.items():
        split_keys = list(filter(
            lambda x: x != '', keys.split(separator)
        ))
        cursor = output
        last = split_keys.pop()
        for key in split_keys:
            if key not in cursor:
                cursor[key] = {}

            if not isinstance(cursor[key], dict):
                msg = f"Duplicate key conflict. Key: '{key}'."
                raise KeyError(msg)

            cursor = cursor[key]
        cursor[last] = val
    return output


def unembed(item):
    # type: (Any) -> Any
    '''
    Convert embeded types in dictionary keys into python types.

    Args:
        item (object): Dictionary with embedded types.

    Returns:
        object: Converted object.
    '''
    lut = {'list': list, 'tuple': tuple, 'set': set}
    embed_re = re.compile(r'^<([a-z]+)_(\d+)>$')

    if is_dictlike(item) and item != {}:
        output = {}  # type: Any
        keys = list(item.keys())
        match = embed_re.match(keys[0])
        if match:
            indices = [embed_re.match(key).group(2) for key in keys]  # type: ignore
            indices = map(int, indices)  # type: ignore

            output = []
            for i, key in sorted(zip(indices, keys)):
                next_item = item[key]
                if is_dictlike(next_item):
                    next_item = unembed(next_item)
                output.append(next_item)

            output = lut[match.group(1)](output)
            return output
        else:
            for key, val in item.items():
                output[key] = unembed(val)
        return output
    return item


# FILE-FUNCTIONS----------------------------------------------------------------
def list_all_files(directory):
    # type: (Union[str, Path]) -> List[Path]
    '''
    Recursively lists all files within a give directory.

    Args:
        directory (str or Path): Directory to be recursed.

    Returns:
        list[Path]: List of filepaths.
    '''
    output = []  # type: List[Path]
    for root, dirs, files in os.walk(directory):
        for file_ in files:
            fullpath = Path(root, file_)
            output.append(fullpath)
    return output


def get_parent_fields(key, separator='/'):
    # type: (str, str) -> List[str]
    '''
    Get all the parent fields of a given key, split by given separator.

    Args:
        key (str): Key.
        separator (str, optional): String that splits key into fields.
            Default: '/'.

    Returns:
        list(str): List of absolute parent fields.
    '''
    fields = key.split(separator)
    output = []  # type: List[str]
    for i in range(len(fields) - 1):
        output.append(separator.join(fields[:i + 1]))
    return output


# EXPORT-FUNCTIONS--------------------------------------------------------------
def dot_to_html(dot, layout='dot', as_png=False):
    # type: (pydot.Dot, str, bool) -> Union[HTML, Image]
    '''
    Converts a given pydot graph into a IPython.display.HTML object.
    Used in jupyter lab inline display of graph data.

    Args:
        dot (pydot.Dot): Pydot Graph instance.
        layout (str, optional): Graph layout style.
            Options include: circo, dot, fdp, neato, sfdp, twopi.
            Default: dot.
        as_png (bool, optional): Display graph as a PNG image instead of SVG.
            Useful for display on Github. Default: False.

    Raises:
        ValueError: If invalid layout given.

    Returns:
        IPython.display.HTML: HTML instance.
    '''
    layouts = ['circo', 'dot', 'fdp', 'neato', 'sfdp', 'twopi']
    if layout not in layouts:
        msg = f'Invalid layout value. {layout} not in {layouts}.'
        raise ValueError(msg)

    if as_png:
        return Image(data=dot.create_png())

    svg = dot.create_svg(prog=layout)
    html = f'<object type="image/svg+xml" data="data:image/svg+xml;{svg}"></object>'  # type: Any
    html = HTML(html)
    html.data = re.sub(r'\\n|\\', '', html.data)
    html.data = re.sub('</svg>.*', '</svg>', html.data)
    return html


def write_dot_graph(
    dot,
    fullpath,
    layout='dot',
):
    # type: (pydot.Dot, Union[str, Path], str) -> None
    '''
    Writes a pydot.Dot object to a given filepath.
    Formats supported: svg, dot, png.

    Args:
        dot (pydot.Dot): Pydot Dot instance.
        fulllpath (str or Path): File to be written to.
        layout (str, optional): Graph layout style.
            Options include: circo, dot, fdp, neato, sfdp, twopi. Default: dot.

    Raises:
        ValueError: If invalid file extension given.
    '''
    if isinstance(fullpath, Path):
        fullpath = Path(fullpath).absolute().as_posix()

    _, ext = os.path.splitext(fullpath)
    ext = re.sub(r'^\.', '', ext)
    if re.search('^svg$', ext, re.I):
        dot.write_svg(fullpath, prog=layout)
    elif re.search('^dot$', ext, re.I):
        dot.write_dot(fullpath, prog=layout)
    elif re.search('^png$', ext, re.I):
        dot.write_png(fullpath, prog=layout)
    else:
        msg = f'Invalid extension found: {ext}. '
        msg += 'Valid extensions include: svg, dot, png.'
        raise ValueError(msg)



from typing import Any, Dict, Optional

from flask import Flask, Response, request, redirect, url_for
from flasgger import Swagger, swag_from
import lunchbox.tools as lbt

from rolling_pin.blob_etl import BlobETL
# ------------------------------------------------------------------------------


'''
Rolling-Pin Flask service.
'''


app = Flask(__name__)
swagger = Swagger(app)


@app.route('/')
def index():
    # type: () -> Any
    return redirect(url_for('flasgger.apidocs'))


@lbt.api_function
def get_svg(
    data='<required>',
    layout='dot',
    orthogonal_edges=False,
    orient='tb',
    color_scheme=None,
):
    # type: (str, str, bool, str, Optional[Dict]) -> str
    '''
    Generate a SVG string from a given JSON blob.

    Args:
        data (dict or list): JSON blob.
        layout (str, optional): Graph layout style.
            Options include: circo, dot, fdp, neato, sfdp, twopi. Default: dot.
        orthogonal_edges (bool, optional): Whether graph edges should have
                non-right angles. Default: False.
            orient (str, optional): Graph layout orientation. Default: tb.
                Options include:

                * tb - top to bottom
                * bt - bottom to top
                * lr - left to right
                * rl - right to left
            color_scheme: (dict, optional): Color scheme to be applied to graph.
                Default: rolling_pin.tools.COLOR_SCHEME

    Returns:
        str: SVG string.
    '''
    output = BlobETL(data)\
        .to_dot_graph(
            orthogonal_edges=orthogonal_edges,
            orient=orient,
            color_scheme=color_scheme)\
        .create_svg(prog=layout)\
        .decode('utf-8')
    return output


@app.route('/to_svg')
@swag_from(dict(
    parameters=[
        dict(
            name='data',
            type='object',
            required=True,
        ),
        dict(
            name='layout',
            type='string',
            description='Graph layout style.',
            enum=['circo', 'dot', 'fdp', 'neato', 'sfdp', 'twopi'],
            required=False,
            default='dot',
        ),
        dict(
            name='orthogonal_edges',
            type='boolean',
            required=False,
            description='Whether graph edges should have non-right angles.',
            default=False,
        ),
        dict(
            name='orient',
            type='string',
            required=False,
            description='Graph layout orientation.',
            default='tb',
            enum=['tb', 'bt', 'lr', 'rl'],
        ),
        dict(
            name='color_scheme',
            type='object',
            required=False,
            description='Color scheme to be applied to graph.',
            default='none',
        )
    ],
    responses={
        200: dict(
            description='A SVG image.',
            content='image/svg+xml',
        ),
        400: dict(
            description='Invalid JSON request sent.',
            example=dict(
                error="KeyError: data",
                status=400,
                success=False,
            )
        )
    }
))
def to_svg():
    # type: () -> Response
    '''
    Endpoint for converting a given JSON blob into a SVG graph.
    '''
    try:
        params = request.get_json()
        return Response(get_svg(**params), mimetype='image/svg+xml')
    except Exception as e:
        msg = e.__class__.__name__ + ': ' + ' '.join(e.args)
        return Response(msg, status=400)


if __name__ == '__main__':
    app.run(debug=True, host='0.0.0.0', port=8080)



from setuptools import setup

with open('version.txt') as f:
    VERSION = f.read().strip('\n')

with open('prod_requirements.txt') as f:
    PROD_REQUIREMENTS = f.read().split('\n')

with open('dev_requirements.txt') as f:
    DEV_REQUIREMENTS = f.read().split('\n')

with open('README.md') as f:
    README = f.read()
# ------------------------------------------------------------------------------

setup(
    name='rolling-pin',
    packages=['rolling_pin'],
    version=VERSION,
    license='MIT',
    description='A library of generic tools for ETL work and visualization of JSON blobs and python repositories.',  # noqa E501
    long_description=README,
    long_description_content_type='text/markdown',
    author='Alex Braun',
    author_email='Alexander.G.Braun@gmail.com',
    url='https://github.com/theNewFlesh/rolling-pin',
    download_url='https://github.com/theNewFlesh/rolling-pin/archive/' + VERSION + '.tar.gz',
    keywords=['ETL', 'blob', 'dependency', 'graph', 'svg', 'networkx', 'transform'],
    install_requires=PROD_REQUIREMENTS,
    classifiers=[
      'Development Status :: 4 - Beta',
      'Intended Audience :: Developers',
      'License :: OSI Approved :: MIT License',
      'Programming Language :: Python :: 3',
      'Programming Language :: Python :: 3.6',
      'Programming Language :: Python :: 3.7',
    ],
    extras_require={
        "dev": DEV_REQUIREMENTS
    },
)
